import type { ComponentPublicInstance, ReactiveFlags, CreateComponentPublicInstance, WatchOptions } from "vue";
import { DateKind } from "./metadata";
export type OwnProps<T, TExclude> = Pick<T, Exclude<keyof T, keyof TExclude>>;
export type Indexable<T> = {
    [k: string]: any | undefined;
} & T;
export type DeepPartial<T> = T | {
    [P in keyof T]?: T[P] extends Array<infer U> ? Array<DeepPartial<U>> : T[P] extends ReadonlyArray<infer U> ? ReadonlyArray<DeepPartial<U>> : DeepPartial<T[P]>;
};
export declare function isNullOrWhitespace(value: string | null | undefined): boolean;
export declare function parseJSONDate(argument: string, kind?: DateKind): Date;
export declare function parseDateUserInput(input: string, defaultDate: Date, dateKind: DateKind): Date | undefined;
/**
 * Converts deep Javascript objects to URL encoded query strings.
 * @license MIT
 */
export declare function objectToQueryString(a: Array<any> | {
    [s: string]: any;
} | null): string;
export declare function objectToFormData(a: Array<any> | {
    [s: string]: any;
} | null): FormData;
/** This is the property added by `markRaw`.
 * We use it directly so we can declare it on the proto of ViewModel/ListViewModel
 * rather than calling markRaw on each instance.
 *
 * We make these classes nonreactive with Vue, preventing ViewModel instance from being wrapped with a Proxy.
 * To achieve reactivity, we instead make individual members reactive with `reactive`/`ref`.
 *
 * We have to do this because `reactive` doesn't play nice with prototyped objects.
 * Any sets to a setter on the ViewModel class will trigger the reactive proxy,
 * and since the setter is defined on the prototype and Vue checks hasOwnProperty
 * when determining if a field is new on an object, all setters trigger reactivity
 * even if the value didn't change.
 *
 * We cant use the export of ReactiveFlags from vue because of https://github.com/vuejs/core/issues/1228
 */
export declare const ReactiveFlags_SKIP: ReactiveFlags.SKIP;
type RelevantVueProps = "$route" | "$router" | "$nextTick";
/** A type that accepts a component instance in both Vue2 and Vue3. Notably, the type of `this` in methods in the options API is not assignable to ComponentPublicInstance, but it is to CreateComponentPublicInstance. */
export type VueInstance = Pick<ComponentPublicInstance | CreateComponentPublicInstance, RelevantVueProps> & {
    $watch(expOrFn: string, callback: (n: any, o: any) => void, options?: WatchOptions): () => void;
    $watch<T>(expOrFn: () => T, callback: (n: T, o: T) => void, options?: WatchOptions): () => void;
};
export declare function getInternalInstance(vue: VueInstance): any;
export declare const IsVue2: boolean;
export declare const IsVue3: boolean;
export {};
